// $Id: test3.C,v 1.16 2000/04/24 02:50:58 wylie Exp $
//
// libdyninst validation suite test #3
//    Author: Jeff Hollingsworth (6/18/99)
//

//  This program tests having multiple active mutatee processes.
//   
//  To run a subset of the tests, enter -run <test nums> on the command
//      line.
//
//  Naming conventions:
//      All functions, variables, etc are name funcXX_YY, exprXX_YY, etc.
//          XX is the test number
//          YY is the instance withing the test
//	    func1_2 is the second function used in test case #1.
//

#include <stdio.h>
#include <string.h>
#ifdef i386_unknown_nt4_0
#include <windows.h>
#include <winbase.h>
#define unlink _unlink
#else
#include <unistd.h>
#endif

#include "BPatch.h"
#include "BPatch_Vector.h"
#include "BPatch_thread.h"
#include "BPatch_snippet.h"
#include "test_util.h"

extern "C" const char V_libdyninstAPI[];

int debugPrint = 0; // internal "mutator" tracing
int errorPrint = 0; // external "dyninst" tracing (via errorFunc)

bool runAllTests = true;
const unsigned int MAX_TEST = 4;
bool passedTest[MAX_TEST+1];

template class BPatch_Vector<BPatch_variableExpr*>;

BPatch *bpatch;

static char *mutateeNameRoot = "test3.mutatee";

// control debug printf statements
#define dprintf	if (debugPrint) printf

/**************************************************************************
 * Error callback
 **************************************************************************/

#define DYNINST_NO_ERROR -1

int expectError = DYNINST_NO_ERROR;

void errorFunc(BPatchErrorLevel level, int num, const char **params)
{
    if (num == 0) {
        // conditional reporting of warnings and informational messages
        if (errorPrint) {
            if (level == BPatchInfo)
              { if (errorPrint > 1) printf("%s\n", params[0]); }
            else
                printf("%s", params[0]);
        }
    } else {
        // reporting of actual errors
        char line[256];
        const char *msg = bpatch->getEnglishErrorString(num);
        bpatch->formatErrorString(line, sizeof(line), msg, params);
        
        if (num != expectError) {
            printf("Error #%d (level %d): %s\n", num, level, line);
        
            // We consider some errors fatal.
            if (num == 101) {
               exit(-1);
            }
        }
    }
}

/**************************************************************************
 * Utility functions
 **************************************************************************/

//
// Return a pointer to a string identifying a BPatch_procedureLocation
//
char *locationName(BPatch_procedureLocation l)
{
    switch(l) {
      case BPatch_entry:
	return "entry";
      case BPatch_exit:
	return "exit";
      case BPatch_subroutine:
	return "call points";
      case BPatch_longJump:
	return "long jump";
      case BPatch_allLocations:
	return "all";
      default:
	return "<invalid BPatch_procedureLocation>";
    };
}


//
// Insert "snippet" at the location "loc" in the function "inFunction."
// Returns the value returned by BPatch_thread::insertSnippet.
//
BPatchSnippetHandle *insertSnippetAt(BPatch_thread *appThread,
	BPatch_image *appImage, char *inFunction, BPatch_procedureLocation loc,
	BPatch_snippet &snippet, int testNo, char *testName)
{
    // Find the point(s) we'll be instrumenting
    BPatch_Vector<BPatch_point *> *points =
	appImage->findProcedurePoint(inFunction, loc);

    if (!points) {
	fprintf(stderr, "**Failed** test #%d (%s)\n", testNo, testName);
	fprintf(stderr, "    Unable to find point %s - %s\n",
		inFunction, locationName(loc));
	exit(-1);
    }

    return appThread->insertSnippet(snippet, *points);
}

//
// Create a snippet that calls the function "funcName" with no arguments
//
BPatch_snippet *makeCallSnippet(BPatch_image *appImage, char *funcName,
				int testNo, char *testName)
{
    BPatch_function *call_func = appImage->findFunction(funcName);
    if (call_func == NULL) {
	fprintf(stderr, "**Failed** test #%d (%s)\n", testNo, testName);
	fprintf(stderr, "    Unable to find function %s\n", funcName);
	exit(1);
    }

    BPatch_Vector<BPatch_snippet *> nullArgs;
    BPatch_snippet *ret = new BPatch_funcCallExpr(*call_func, nullArgs);

    if (ret == NULL) {
	fprintf(stderr, "**Failed** test #%d (%s)\n", testNo, testName);
	fprintf(stderr, "    Unable to create snippet to call %s\n", funcName);
	exit(1);
    }

    return ret;
}

//
// Insert a snippet to call function "funcName" with no arguments into the
// procedure "inFunction" at the points given by "loc."
//
BPatchSnippetHandle *insertCallSnippetAt(BPatch_thread *appThread,
	BPatch_image *appImage, char *inFunction, BPatch_procedureLocation loc,
	char *funcName, int testNo, char *testName)
{
    BPatch_snippet *call_expr =
	makeCallSnippet(appImage, funcName, testNo, testName);

    BPatchSnippetHandle *ret = insertSnippetAt(appThread, appImage,
					       inFunction, loc, *call_expr,
					       testNo, testName);
    if (ret == NULL) {
	fprintf(stderr, "**Failed** test #%d (%s)\n", testNo, testName);
	fprintf(stderr, "    Unable to insert snippet to call function %s\n",
		funcName);
	exit(-1);
    }

    delete call_expr;
    
    return ret;
}


/**************************************************************************
 * Tests
 **************************************************************************/

//
// Start Test Case #1 - create two processes and process events from each
//     Just let them run to finish, no instrumentation added.
//
void mutatorTest1(char *pathname, BPatch *bpatch)
{
    char *child_argv[4];

    child_argv[0] = pathname;
    child_argv[1] = "-?";
    child_argv[2] = NULL;
    BPatch_thread *appThread1, *appThread2;

    // Start the 1st mutatee
    dprintf("Starting \"%s\" #1\n", pathname);
    child_argv[1] = "1";		// run test1 in mutatee
    appThread1 = bpatch->createProcess(pathname, child_argv, NULL);
    if (!appThread1) {
	printf("*ERROR*: unable to create handle1 for executable\n");
        return;
    }

    // Start the 2nd mutatee
    dprintf("Starting \"%s\" #2\n", pathname);
    child_argv[1] = "1";		// run test1 in mutatee 
    appThread2 = bpatch->createProcess(pathname, child_argv, NULL);
    if (!appThread2) {
	printf("*ERROR*: unable to create handle2 for executable\n");
        return;
    }

    appThread1->continueExecution();
    appThread2->continueExecution();

    while (!appThread1->isTerminated() || !appThread2->isTerminated())
	bpatch->waitForStatusChange();

    if (appThread1->isTerminated() && appThread2->isTerminated()) {
	printf("Passed Test #1\n");
	passedTest[1] = true;
    }
}

//
// read the result code written to the file test3.out.<pid> and return it
//
int readResult(int pid)
{
    int ret;
    FILE *fp;
    char filename[80];

    sprintf(filename, "test3.out.%d", pid);
    fp = fopen(filename, "r");
    if (!fp) {
	printf("ERROR: unable to open output file %s\n", filename);
	return -1;
    }
    fscanf(fp, "%d\n", &ret);
    fclose(fp);
    // don't need the file any longer so delete it now
    unlink(filename);

    return ret;
}

//
// Start Test Case #2 - create two processes and insert different code into
//     each one.  The code sets a global variable which the mutatee then
//     writes to a file.  After both mutatee exit, the mutator reads the
//     files to verify that the correct code ran in each mutatee.
//     The first mutator should write a 1 to the file and the second a 2.
//     If no code is patched into the mutatees, the value is 0xdeadbeef.
//
void mutatorTest2(char *pathname, BPatch *bpatch)
{
    char *child_argv[4];

    int pid1, pid2;
    child_argv[0] = pathname;
    child_argv[1] = "-?";
    child_argv[2] = NULL;
    BPatch_thread *appThread1, *appThread2;

    // Start the 1st mutatee
    dprintf("Starting \"%s\" #1\n", pathname);
    child_argv[1] = "2";		// run test2 in mutatee
    appThread1 = bpatch->createProcess(pathname, child_argv, NULL);
    if (!appThread1) {
	printf("*ERROR*: unable to create handle1 for executable\n");
        return;
    }
    pid1 = appThread1->getPid();
    // Start the 2nd mutatee
    dprintf("Starting \"%s\" #2\n", pathname);
    child_argv[1] = "2";		// run test2 in mutatee 
    appThread2 = bpatch->createProcess(pathname, child_argv, NULL);
    if (!appThread2) {
	printf("*ERROR*: unable to create handle2 for executable\n");
        return;
    }
    pid2 = appThread2->getPid();

    BPatch_image *img1 = appThread1->getImage();
    BPatch_image *img2 = appThread2->getImage();

    BPatch_variableExpr *var1 = img1->findVariable("test2ret");
    BPatch_arithExpr snip1(BPatch_assign, *var1, BPatch_constExpr(1));
    insertSnippetAt(appThread1, img1, "func2_1", BPatch_entry, snip1, 2,
	"different inst. active");

    BPatch_variableExpr *var2 = img2->findVariable("test2ret");
    BPatch_arithExpr snip2(BPatch_assign, *var2, BPatch_constExpr(2));
    insertSnippetAt(appThread2, img2, "func2_1", BPatch_entry, snip2, 2,
	"different inst. active");

    appThread1->continueExecution();
    appThread2->continueExecution();

    while (!appThread1->isTerminated() || !appThread2->isTerminated())
	bpatch->waitForStatusChange();

    // now read the files to see if the value is what is expected
    int ret1 = readResult(pid1);
    int ret2 = readResult(pid2);

    if ((ret1 != 1) || (ret2 != 2)) {
	printf("**Failed** test case #2\n");
	if (ret1 != 1) printf("    1st process produced %d, not 1\n", ret1);
	if (ret2 != 2) printf("    2nd process produced %d, not 2\n", ret2);
    } else {
	printf("Passed Test #2\n");
	passedTest[2] = true;
    }
}

//
// Start Test Case #3 - create one process, wait for it to exit.  Then 
//     create a second one and wait for it to exit.
//
void mutatorTest3(char *pathname, BPatch *bpatch)
{
    char *child_argv[4];

    child_argv[0] = pathname;
    child_argv[1] = "-?";
    child_argv[2] = NULL;
    BPatch_thread *appThread1, *appThread2;

    // Start the 1st mutatee
    dprintf("Starting \"%s\" #1\n", pathname);
    child_argv[1] = "1";		// run test1 in mutatee
    appThread1 = bpatch->createProcess(pathname, child_argv, NULL);
    if (!appThread1) {
	printf("*ERROR*: unable to create handle1 for executable\n");
        return;
    }

    appThread1->continueExecution();

    while (!appThread1->isTerminated())
	bpatch->waitForStatusChange();

    // Start the 2nd mutatee
    dprintf("Starting \"%s\" #2\n", pathname);
    child_argv[1] = "1";		// run test1 in mutatee 
    appThread2 = bpatch->createProcess(pathname, child_argv, NULL);
    if (!appThread2) {
	printf("*ERROR*: unable to create handle2 for executable\n");
        return;
    }

    appThread2->continueExecution();
    while (!appThread2->isTerminated())
	bpatch->waitForStatusChange();

    printf("Passed Test #3\n");
    passedTest[3] = true;
}


//
// Start Test Case #4 - create one process, wait for it to exit.  Then 
//     create a second one and wait for it to exit.  This differs from test 3
//     in that the mutatee processes terminate with abort rather than exit.
//
void mutatorTest4(char *pathname, BPatch *bpatch)
{
    char *child_argv[4];

    child_argv[0] = pathname;
    child_argv[1] = "-?";
    child_argv[2] = NULL;
    BPatch_thread *appThread1, *appThread2;

    // Start the 1st mutatee
    dprintf("Starting \"%s\" #1\n", pathname);
    child_argv[1] = "4";		// run test4 in mutatee
    appThread1 = bpatch->createProcess(pathname, child_argv, NULL);
    if (!appThread1) {
	printf("*ERROR*: unable to create handle1 for executable\n");
        return;
    }

    appThread1->continueExecution();

    while (!appThread1->isTerminated())
	bpatch->waitForStatusChange();

    // Start the 2nd mutatee
    dprintf("Starting \"%s\" #2\n", pathname);
    child_argv[1] = "4";		// run test4 in mutatee 
    appThread2 = bpatch->createProcess(pathname, child_argv, NULL);
    if (!appThread2) {
	printf("*ERROR*: unable to create handle2 for executable\n");
        return;
    }

    appThread2->continueExecution();
    while (!appThread2->isTerminated())
	bpatch->waitForStatusChange();

    printf("Passed Test #4\n");
    passedTest[4] = true;
}

int main(unsigned int argc, char *argv[])
{
    bool N32ABI=false;
    char mutateeName[128];
    char libRTname[256];
    bool runTest[MAX_TEST+1];		// should we run a particular test

    strcpy(mutateeName,mutateeNameRoot);
    libRTname[0]='\0';

#if !defined(USES_LIBDYNINSTRT_SO)
    fprintf(stderr,"(Expecting subject application to be statically linked"
                        " with libdyninstAPI_RT.)\n");
#else
    if (!getenv("DYNINSTAPI_RT_LIB")) {
	 fprintf(stderr,"Environment variable DYNINSTAPI_RT_LIB undefined:\n"
#if defined(i386_unknown_nt4_0)
		 "    using standard search strategy for libdyninstAPI_RT.dll\n");
#else
	         "    set it to the full pathname of libdyninstAPI_RT\n");   
         exit(-1);
#endif
    } else
         strcpy((char *)libRTname, (char *)getenv("DYNINSTAPI_RT_LIB"));
#endif

    unsigned int i;
    // by default run all tests
    for (i=1; i <= MAX_TEST; i++) {
	runTest[i] = true;
	passedTest[i] = false;
    }

    for (i=1; i < argc; i++) {
        if (strncmp(argv[i], "-v+", 3) == 0)    errorPrint++;
        if (strncmp(argv[i], "-v++", 4) == 0)   errorPrint++;
	if (strncmp(argv[i], "-verbose", 2) == 0) {
	    debugPrint = 1;
	} else if (!strcmp(argv[i], "-V")) {
            fprintf (stdout, "%s\n", V_libdyninstAPI);
            if (libRTname[0]) 
                fprintf (stdout, "DYNINSTAPI_RT_LIB=%s\n", libRTname);
            fflush(stdout);
        } else if (!strcmp(argv[i], "-skip")) {
            unsigned int j;
            runAllTests = false;
            for (j=i+1; j < argc; j++) {
                unsigned int testId;
                if ((testId = atoi(argv[j]))) {
                    if ((testId > 0) && (testId <= MAX_TEST)) {
                        runTest[testId] = false;
                    } else {
                        printf("invalid test %d requested\n", testId);
                        exit(-1);
                    }
                } else {
                    // end of test list
                    break;
                }
            }
            i=j-1;
	} else if (!strcmp(argv[i], "-run")) {
	    unsigned int j;
            runAllTests = false;
	    for (j=0; j <= MAX_TEST; j++) runTest[j] = false;
	    for (j=i; j < argc; j++) { 
		unsigned int testId;
		if ((testId = atoi(argv[j]))) {
		    if ((testId > 0) && (testId <= MAX_TEST)) {
			runTest[testId] = true;
		    } else {
			printf("invalid test %d requested\n", testId);
			exit(-1);
		    }
		} else {
		    // end of test list
		}
	    }
	    i=j-1;
	} else if (!strcmp(argv[i], "-mutatee")) {
	    i++;
            if (*argv[i]=='_')
                strcat(mutateeName,argv[i]);
            else
                strcpy(mutateeName,argv[i]);
#if defined(mips_sgi_irix6_4)
	} else if (!strcmp(argv[i], "-n32")) {
	    N32ABI=true;
#endif
	} else {
	    fprintf(stderr, "Usage: test3 "
		    "[-V] [-verbose] "
#if defined(mips_sgi_irix6_4)
		    "[-n32] "
#endif
                    "[-mutatee <test3.mutatee>] "
                    "[-run <test#> <test#> ...] "
                    "[-skip <test#> <test#> ...]\n");
	    exit(-1);
        }
    }

    if (!runAllTests) {
        printf("Running Tests: ");
	for (unsigned int j=1; j <= MAX_TEST; j++) {
	    if (runTest[j]) printf("%d ", j);
	}
	printf("\n");
    }

    // patch up the default compiler in mutatee name (if necessary)
    if (!strstr(mutateeName, "_"))
#if defined(i386_unknown_nt4_0)
        strcat(mutateeName,"_VC");
#else
        strcat(mutateeName,"_gcc");
#endif
    if (N32ABI || strstr(mutateeName,"_n32")) {
        // patch up file names based on alternate ABI (as necessary)
        if (!strstr(mutateeName, "_n32")) strcat(mutateeName,"_n32");
    }
    // patch up the platform-specific filename extensions
#if defined(i386_unknown_nt4_0)
    if (!strstr(mutateeName, ".exe")) strcat(mutateeName,".exe");
#endif

    // Create an instance of the BPatch library
    bpatch = new BPatch;

#if defined (sparc_sun_solaris2_4)
    // we use some unsafe type operations in the test cases.
    bpatch->setTypeChecking(false);
#endif

    // Register a callback function that prints any error messages
    bpatch->registerErrorCallback(errorFunc);

    if (runTest[1]) mutatorTest1(mutateeName, bpatch);
    if (runTest[2]) mutatorTest2(mutateeName, bpatch);
    if (runTest[3]) mutatorTest3(mutateeName, bpatch);
    if (runTest[4]) mutatorTest4(mutateeName, bpatch);

    unsigned int testsFailed = 0;
    for (i=1; i <= MAX_TEST; i++) {
	if (runTest[i] && !passedTest[i]) testsFailed++;
    }

    if (!testsFailed) {
	if (runAllTests) {
	    printf("All tests passed\n");
	} else {
	    printf("All requested tests passed\n");
	}
    } else {
	printf("**Failed** %d test%c\n",testsFailed,(testsFailed>1)?'s':' ');
    }

    return (testsFailed ? 127 : 0);
}
